﻿using System;
using System.Linq;
using System.Net;
using System.Net.Sockets;
using System.Security.Cryptography;
using System.Text;

namespace UdPunching
{
    public static class UtilMethods
    {
        private static readonly DateTime UnixTimeStampOrig = new DateTime(1970, 1, 1);

        public static bool IpEndPointEqualsTo(this EndPoint me, EndPoint another)
        {
            if (me is IPEndPoint ipMe && another is IPEndPoint ipAnother)
                return ipMe.Port == ipAnother.Port && ipMe.Address.Equals(ipAnother.Address);
            return false;
        }

        public static int ToUnixTimeStamp(this DateTime me)
        {
            return (int)me.ToUniversalTime().Subtract(UnixTimeStampOrig).TotalSeconds;
        }

        public static IPEndPoint ParseToIpEndPointV4(this string str)
        {
            var parts = str.Split(':');
            if (2 != parts.Length) throw new ArgumentException("invalid addr port parts");
            if (IPAddress.TryParse(parts[0], out var addr) && int.TryParse(parts[1], out var port)) return new IPEndPoint(addr, port);
            throw new FormatException("parse fail");
        }

        public static void SendExchangeMessageTo(this Socket socket, Guid senderId, ExchangeMessage message, EndPoint to, RSACryptoServiceProvider receiverKey)
        {
            var bytes = TransferCodec.Encode(receiverKey, senderId, message.ToBytes());
            socket.SendTo(bytes, to);
        }

        //========= underlying =============

        public static string ToPaddingString(this IPAddress ipAddress)
        {
            if (ipAddress.AddressFamily != AddressFamily.InterNetwork) throw new NotSupportedException("Only ipv4");
            return ipAddress.ToString().PadRight(15, ' ');
        }

        //--------- ToBytes Series -------------

        public static byte[] ToLeInt16Bytes(this int value)
        {
            return new[]
            {
                (byte)(value & 0xFF),
                (byte)((value>>8 ) & 0xFF),
            };
        }

        public static byte[] ToLeInt32Bytes(this int value)
        {
            return new[]
            {
                (byte)(value & 0xFF),
                (byte)((value>>8 ) & 0xFF),
                (byte)((value>>16 ) & 0xFF),
                (byte)((value>>24 ) & 0xFF),
            };
        }

        public static byte[] ToAsciiBytes(this string str) => Encoding.ASCII.GetBytes(str);

        //--------- Read Series -------------

        public static int ReadLeInt16(this byte[] buf, out int value, int index = 0)
        {
            value = buf[index]
                 | buf[index + 1] << 8
            ;

            return 2;
        }

        public static int ReadLeInt16(this byte[] buf, int index = 0)
        {
            return buf[index]
                 | buf[index + 1] << 8
            ;
        }

        public static int ReadLeInt32(this byte[] buf, int index = 0)
        {
            return buf[index]
                | (buf[index + 1] << 8)
                | (buf[index + 2] << 16)
                | (buf[index + 3] << 24)
            ;
        }

        public static int ReadUnixTimeStamp(this byte[] buf, int index, out DateTime timeStamp)
        {
            var i32 = buf.ReadLeInt32(index);
            timeStamp = UnixTimeStampOrig.AddSeconds(i32).ToLocalTime();
            return 4;
        }

        public static int ReadIpEndPoint(this byte[] buf, int index, out IPEndPoint ipEndPoint)
        {
            index += buf.ReadAsciiString(15, out var strAddress, index);
            buf.ReadLeInt16(out var port, index);

            ipEndPoint = new IPEndPoint(IPAddress.Parse(strAddress.Trim()), port);
            return 17;
        }

        public static int ReadAsciiString(this byte[] buf, int count, out string ascii, int index = 0)
        {
            ascii = Encoding.ASCII.GetString(buf.Skip(index).Take(count).ToArray());
            return count;
        }

        //--------- Write...To Series -------------

        public static int WriteLeInt16To(this int value, byte[] buf, int index = 0)
        {
            buf[index] = (byte)(value & 0xFF);
            buf[index + 1] = (byte)((value >> 8) & 0xFF);
            return 2;
        }

        public static int WriteTo(this Guid value, byte[] buf, int index = 0)
        {
            var byteArray = value.ToByteArray();
            byteArray.CopyTo(buf, index);
            return byteArray.Length;
        }

        public static int WriteTo(this byte[] bytes, byte[] buf, int index = 0)
        {
            bytes.CopyTo(buf, index);
            return bytes.Length;
        }

        public static int WriteAsciiBytesTo(this string str, byte[] buf, int index = 0)
        {
            Encoding.ASCII.GetBytes(str, 0, str.Length, buf, index);
            return Encoding.ASCII.GetByteCount(str);
        }
    }
}